﻿// Hangeul Typing Behavior Analyzer 2.0
// created by Kim Yongmook ( http://moogi.new21.org )
// Originically created on March 7, 2000
// Modified on April 21, 2005
// Ver 2.0 released by Pat-al ( http://pat.im ) on November 27, 2011 
// Ver 2.01 released by Pat-al on December 11, 2011

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <io.h>
#include <errno.h>

//Member: 0->Shift?, 1->Finger, 2->Height in keyboard, 3->Left/Right relative key position
char typeinfo[128][4]={{0,}};

#define sett(n,a,b,c,d) typeinfo[n][0]=a, typeinfo[n][1]=b, typeinfo[n][2]=c, typeinfo[n][3]=d;
// n:문자, a:shift(0/1), b:손가락+- c:줄(아래부터 1~5) d:좌우 상대 위치
#define is_hanja(ch1,ch2) \
	( ch1>=0xE0 && ch1<=0xF9 && ((ch2>=0x31 && ch2<=0x7E) || (ch2>0x91 && ch2<=0xFE)) )
 
int *key_untreated_character_total;		// 자판 배열에 없어서 처리하지 못한 문자
int *key_set_type;						// 벌식(2,3)
char **key_fingering;					// 운지법
char **key_layout_filenames;			// 자판 배열 파일 이름

void init_type_info(int key_layout_num)	// set_type:벌식 구분, fingering:운지법 구분(pyojun,gong)
{
	int i;
	//sett('문자', shift, 손가락, 줄, 좌우 상대 위치)

	sett('a',0,-5,3,-5); sett('s',0,-4,3,-4); sett('d',0,-3,3,-3); sett('f',0,-2,3,-2); sett('g',0,-2,3,-1);
	sett('h',0,2,3,1); sett('j',0,2,3,2); sett('k',0,3,3,3); sett('l',0,4,3,4);

	sett('q',0,-5,4,-5); sett('w',0,-4,4,-4); sett('e',0,-3,4,-3); sett('r',0,-2,4,-2); sett('t',0,-2,4,-1);
	sett('y',0,2,4,1); sett('u',0,2,4,2); sett('i',0,3,4,3); sett('o',0,4,4,4); sett('p',0,5,4,5);
	sett('z',0,-5,2,-5); sett('x',0,-4,2,-4); sett('c',0,-3,2,-3); sett('v',0,-2,2,-2);	// 표준식 왼쪽 밑글쇠 치기
	if(!stricmp(key_fingering[key_layout_num],"Gong") || !stricmp(key_fingering[key_layout_num],"Gong3")) {
		sett('z',0,-4,2,-4); sett('x',0,-3,2,-3); sett('c',0,-2,2,-2); sett('v',0,-2,2,-2);	// 공병우 자판식 왼쪽 밑글쇠 치기 (공 운지법)
	}

	sett('b',0,-2,2,-1);
	if(key_set_type[key_layout_num]==2 && !stricmp(key_fingering[key_layout_num],"Pyojun")) sett('b',0,2,2,1);	// 'ㅠ'(B 자리)를 오른손으로 치기 (두벌 자판)
	sett('n',0,2,2,1); sett('m',0,2,2,2);
	for(i='A';i<='Z';i++) sett(i,1,typeinfo[i+32][1],typeinfo[i+32][2],typeinfo[i+32][3]);

	sett('1',0,-5,5,-5); sett('2',0,-4,5,-4); sett('3',0,-3,5,-3); sett('4',0,-2,5,-2); sett('5',0,-2,5,-1);
	sett('6',0,2,5,1);	sett('7',0,2,5,2); sett('8',0,3,5,3); sett('9',0,4,5,4); sett('0',0,5,5,5);

	sett('!',1,-5,5,-5); sett('@',1,-4,5,-4); sett('#',1,-3,5,-3); sett('$',1,-2,5,-2); sett('%',1,-2,5,-1);
	sett('^',1,2,5,1); sett('&',1,2,5,2); sett('*',1,3,5,3); sett('(',1,4,5,4); sett(')',1,5,5,5);

	if(key_set_type[key_layout_num]==3 && !stricmp(key_fingering[key_layout_num],"Gong3")) {	// 공병우 계열 3벌 자판은 'ᅣ'(6 자리) 글쇠를 왼손으로 침
		sett('6',0,-2,5,1); sett('^',1,-2,5,1);
	}
 
	sett('`',0,-5,5,-6); sett('-',0,5,5,6); sett('=',0,5,5,7);
	sett('~',1,-5,5,-6); sett('_',1,5,5,6); sett('+',1,5,5,7);
	sett('[',0,5,4,6); sett(']',0,5,4,7); sett('\\',0,5,4,8);
	sett('{',1,5,4,6); sett('}',1,5,4,7); sett('|',1,5,4,8);
	sett(';',0,5,3,5); sett('\'',0,5,3,6);
	sett(':',1,5,3,5); sett('\"',1,5,3,6);
	sett(',',0,3,2,3); sett('.',0,4,2,4); sett('/',0,5,2,5);
	sett('<',1,3,2,3); sett('>',1,4,2,4); sett('?',1,5,2,5);

	sett('\n',0,5,3,7); sett(' ',0,1,1,0);
	//sett('문자',shift,'-+손가락',줄,좌우 상대 위치)   sett(n,a,b,c) typeinfo[n][0]=a, typeinfo[n][1]=b, typeinfo[n][2]=c;
}

void parse(char *keyseq, int key_layout_num)
{
	int i,j,k, length;
	int counted_enter=0, counted_space=0;	// 줄바꿈, 빈 칸을 셈한 수
	int left_general_key_stroke=0, right_general_key_stroke=0;	// 일반 글쇠 누르는 수
	int left_shift_stroke=0, right_shift_stroke=0;	// shift 누르는 새끼 손가락 치는 수
	int total_general_key_stroke=0;

	int prev_hand=0, prev_finger=-1, prev_line=-1; //앞에 누른 글쇠 상태
	int hand, finger, finger_position, line; // 지금 누르는 글쇠 상태: hand - 누른 손, finger - 누른 손가락

	int shift_hand=false, shift_finger=-1; // shift 누른 글쇠, 손
	int prev_shift_hand=false, prev_shift_finger=-1; // shift 누른 글쇠, 손

	int total_shift_stroke=0;// 윗글쇠 누른 수, 같은 글쇠 거듭치기
	int left_shift_same_finger_stroke=0, right_shift_same_finger_stroke=0; // 손가락 거듭치기에 들어간 윗글쇠 타수
	int left_same_finger_stroke=0, right_same_finger_stroke=0;	// 같은 손가락 거듭치기
	int total_same_key_stroke=0, left_same_key_stroke=0, right_same_key_stroke=0; // 같은 글쇠 거듭치기
	int left_same_hand_stroke_state=0, right_same_hand_stroke_state=0;		// 같은 손으로 거듭 친 수(상태 변수)
	int left_hand_stroke_state=0, right_hand_stroke_state=0;	// 한 손에서 거듭 이어치는 타수

	int left_finger, right_finger; // 치는 손가락 번호
	int same_finger=-1, shift_same_finger_stroke_finger=-1;	// 일반 글쇠와 윗글쇠에서 거듭치기가 일어나는 손가락
	int prev_same_finger=-1; 	// 거듭치기가 일어났던 손가락

	int finger_stroke_distrib[10]={0,};	// 손가락 쓴 수 (0:왼손 새끼, 9:오른손 새끼)
	int finger_line_freq[10][6]={{0,}}; // 손가락(finger), 줄(line)에 따른 글쇠 빈도
	int general_key_position_freq[10][6]={{0,}}; // 손가락(finger), 줄(line)에 따른 글쇠 빈도
	int same_finger_stroke_distrib[10]={0,};	// 각 손가락에서 거듭치기가 일어난 수

	int **continuous_same_hand_stroke;	// 각 손가락에서 거듭치기가 일어난 빈도 수 [왼손 0, 오른손 1][거듭 친 수-1]
	int **continuous_same_finger_stroke; // 각 손가락에서 거듭치기가 일어난 빈도 수 [손가락 0~9][거듭 친 수-1]
	int **continuous_same_key_stroke; // 각 손가락에서 거듭치기가 일어난 빈도 수 [손가락 0~9][거듭 친 수-1]
	continuous_same_hand_stroke = (int **)malloc(2*sizeof(int *));
	continuous_same_finger_stroke = (int **)malloc(10*sizeof(int *));
	continuous_same_key_stroke = (int **)malloc(10*sizeof(int *));
	for(i=0;i<2;++i) continuous_same_hand_stroke[i] = (int *)calloc(100, sizeof(int));
	for(i=0;i<10;++i) continuous_same_finger_stroke[i] = (int *)calloc(100, sizeof(int));
	for(i=0;i<10;++i) continuous_same_key_stroke[i] = (int *)calloc(100, sizeof(int));

	for(i=0,length=strlen(keyseq);i<length;i++) {
		if(i && (keyseq[i]==' ' || keyseq[i]=='\n') && keyseq[i]==keyseq[i-1]) continue; // 거듭 나오는 빈칸이나 빈줄은 하나만 셈한다.
		if(keyseq[i] == 0x18) {
			prev_hand=0; prev_finger=-1;
			prev_line=-1;
			prev_shift_hand=0; prev_shift_finger=-1;
			continue;	// 자판 배열에 없는 문자는 건너뛴다.
		}
		if(keyseq[i] == ' ') ++counted_space;
		if(keyseq[i] == '\n') ++counted_enter;

		hand = typeinfo[keyseq[i]][1]<0 ? -1:1;	// 왼손(-) 오른손(+)
		//if(keyseq[i]=='\n' || keyseq[i]==' ') hand=0;	// 같은 손 거듭치기에 줄바꾸개, 사이띄개를 셈하지 않는다.
		
		finger = typeinfo[keyseq[i]][1]<0 ? typeinfo[keyseq[i]][1]+5 : typeinfo[keyseq[i]][1]+4;	// 왼-오른쪽을 구분하는 손가락 번호 (왼손 다섯째:0, 왼손 첫째:4, 오른손 첫째:5, 오른손 다섯째:9)
		finger_position = typeinfo[keyseq[i]][3]<0 ? typeinfo[keyseq[i]][3]+5 : typeinfo[keyseq[i]][3]+4;	// 첫째 손가락을 빼고, 둘째 손가락 자리를 더 구분하는 손가락 자리 번호

		if(finger_position<0) finger_position=0;
		if(finger_position>9) finger_position=9;
		line = typeinfo[keyseq[i]][2];

		if(keyseq[i] != '\n') ++general_key_position_freq[finger_position][line];	// 줄바꾸개를 빼고 둘째 손가락 자리를 더 구분하는 일반 글쇠의 위아래/손가락별 타수를 더한다.
		++finger_stroke_distrib[finger];		// 손가락별 타수를 셈한다.
		++finger_line_freq[finger][line];		// 손가락/위아래 위치 타수를 셈한다.
		
		++total_general_key_stroke;
		if(keyseq[i] != ' ' && keyseq[i] != '\n') {
			if(hand<0) left_general_key_stroke++;
			if(hand>0) right_general_key_stroke++;
		}

		if(typeinfo[keyseq[i]][0]) {	// shift를 누를 때
			shift_hand = -hand;						// shift를 누르는 손
			shift_finger = shift_hand==1 ? 9:0;		// shift를 누르는 새끼 손가락
			total_shift_stroke++;								// shift 누르는 횟수
			if(shift_hand<0) {						// 일반 글쇠 맞은편의 shift를 치는 것을 반영한다.
				++left_shift_stroke; ++finger_stroke_distrib[0]; ++finger_line_freq[0][2]; /*general_key_position_freq[0][2];*/ 
			} else {
				++right_shift_stroke; ++finger_stroke_distrib[9]; ++finger_line_freq[9][2]; /*general_key_position_freq[9][2];*/ 
			}
		} else {
			shift_hand = false;
			shift_finger = -1;
		}

		same_finger=-1;	left_finger=-1, right_finger=-1;

		if(prev_finger==finger) {	// 일반 글쇠를 한 손가락으로 거듭 칠 때
			same_finger=finger;
			if(hand<0) {
				++left_same_hand_stroke_state; ++left_same_finger_stroke; left_finger=finger; ++same_finger_stroke_distrib[finger];
			} else {
				++right_same_hand_stroke_state; ++right_same_finger_stroke; right_finger=finger; ++same_finger_stroke_distrib[finger];
			}
			if(/*keyseq[i]==keyseq[i-1] &&*/ typeinfo[keyseq[i]][2]==typeinfo[keyseq[i-1]][2] && typeinfo[keyseq[i]][3]==typeinfo[keyseq[i-1]][3]) {
				total_same_key_stroke++;  // 같은 글쇠 거듭치기
				if(hand<0) {	// 왼손으로 칠 때
					for(j=0;j<left_same_hand_stroke_state;++j) if(typeinfo[keyseq[i]][2]!=typeinfo[keyseq[i-j]][2] || typeinfo[keyseq[i]][3]!=typeinfo[keyseq[i-j]][3]) break;
					++left_same_key_stroke;
					++continuous_same_key_stroke[left_finger][j];
					if(left_same_hand_stroke_state>1) --continuous_same_key_stroke[left_finger][j-1];
				} else {		// 오른손으로 칠 때
					for(j=0;j<right_same_hand_stroke_state;++j) if(typeinfo[keyseq[i]][2]!=typeinfo[keyseq[i-j]][2] || typeinfo[keyseq[i]][3]!=typeinfo[keyseq[i-j]][3]) break;
					++right_same_key_stroke;
					++continuous_same_key_stroke[right_finger][j];
					if(right_same_hand_stroke_state>1) --continuous_same_key_stroke[right_finger][j-1];
				}
			}

		}
		if(prev_shift_hand && finger==prev_shift_finger) {	// 윗글쇠를 눌렀던 손가락으로 일반 글쇠를 누를 때
			if(prev_shift_hand<0) left_shift_same_finger_stroke++; else right_shift_same_finger_stroke++;
			if(hand<0) {
				++left_same_hand_stroke_state; ++left_same_finger_stroke; left_finger=finger; ++same_finger_stroke_distrib[finger];
			} else {
				++right_same_hand_stroke_state; ++right_same_finger_stroke; right_finger=finger; ++same_finger_stroke_distrib[finger];				
			}
		}
		if(shift_hand && prev_finger==shift_finger) {	// 일반 글쇠를 눌렀던 손가락으로 윗글쇠를 누를 때
			if(shift_hand<0) {
				left_shift_same_finger_stroke++;
				++left_same_hand_stroke_state; ++left_same_finger_stroke; left_finger=prev_finger; ++same_finger_stroke_distrib[prev_finger];
			} else {
				right_shift_same_finger_stroke++;
				++right_same_hand_stroke_state; ++right_same_finger_stroke; right_finger=prev_finger; ++same_finger_stroke_distrib[prev_finger];
			}
		}

		if(shift_hand && prev_shift_hand && shift_hand==prev_shift_hand && shift_finger==prev_shift_finger) {
			--total_shift_stroke;	// 윗글쇠를 누르고 있을 때는 윗글쇠 타수를 더하지 않는다.
			if(shift_hand<0) {
				--finger_stroke_distrib[0]; --finger_line_freq[0][2]; --left_shift_stroke;
			} else {
				--finger_stroke_distrib[9]; --finger_line_freq[9][2]; --right_shift_stroke;
			}
		}

		if(hand<0 || (shift_hand<0 && shift_hand != prev_shift_hand)) {		// 왼손으로 칠 때 왼손 타수를 셈한다.
			++continuous_same_hand_stroke[0][left_hand_stroke_state];
			if(left_hand_stroke_state>0) --continuous_same_hand_stroke[0][left_hand_stroke_state-1];
			++left_hand_stroke_state;
		} else left_hand_stroke_state = 0;
			
		if(hand>0 || (shift_hand>0 && shift_hand != prev_shift_hand)) {		// 오른손으로 칠 때 오른손 타수를 셈한다.
			++continuous_same_hand_stroke[1][right_hand_stroke_state];	
			if(right_hand_stroke_state>0) --continuous_same_hand_stroke[1][right_hand_stroke_state-1];
			++right_hand_stroke_state;
		} else right_hand_stroke_state = 0;

		if(left_finger == -1) {	
			left_same_hand_stroke_state=0;	//  왼손으로 치지 않았다면 왼손 거듭치기 상태 변수를 초기값으로 돌려 놓는다.
		} else {
			++continuous_same_finger_stroke[left_finger][left_same_hand_stroke_state];
			if(left_same_hand_stroke_state>1) --continuous_same_finger_stroke[left_finger][left_same_hand_stroke_state-1];
		}

		if(right_finger == -1) {	
			right_same_hand_stroke_state=0;	//  오른손으로 치지 않았다면 오른손 거듭치기 상태 변수를 초기값으로 돌려 놓는다.
		} else {
			++continuous_same_finger_stroke[right_finger][right_same_hand_stroke_state];
			if(right_same_hand_stroke_state>1) --continuous_same_finger_stroke[right_finger][right_same_hand_stroke_state-1];
		}
		
		prev_hand=hand; prev_finger=finger; prev_line=typeinfo[keyseq[i]][2];
		prev_shift_hand=shift_hand;
		prev_shift_finger=shift_finger;
	}
	
	int left_keystroke = left_general_key_stroke + left_shift_stroke;
	int right_keystroke = right_general_key_stroke + right_shift_stroke + counted_enter;
	printf("General-key stokes(L/R): %d (%d/%d) (%2.1f/%2.1f%%)\n",total_general_key_stroke ,left_general_key_stroke, right_general_key_stroke, (left_general_key_stroke*100.0f)/(left_general_key_stroke+right_general_key_stroke), (right_general_key_stroke*100.0f)/(left_general_key_stroke+right_general_key_stroke));
	printf("Shift-key strokes(L/R): %d (%d/%d) (%2.1f/%2.1f%%)\n",total_shift_stroke,left_shift_stroke, right_shift_stroke, (left_shift_stroke*100.0f)/(left_shift_stroke+right_shift_stroke), (right_shift_stroke*100.0f)/(left_shift_stroke+right_shift_stroke));
	printf("Counted enter strokes : %d\n",counted_enter, counted_space);
	printf("Total keystrokes except space (L/R): %d (%d/%d) (%2.1f/%2.1f%%)\n",total_general_key_stroke+total_shift_stroke, left_keystroke, right_keystroke, (left_keystroke*100.0f)/(left_keystroke+right_keystroke), (right_keystroke*100.0f)/(left_keystroke+right_keystroke));
	int total_finger=0;
	for(i=0;i<10;++i) if(i!=4 && i!=5) total_finger += finger_stroke_distrib[i];
	printf(" L: "); for(i=3;i>-1;--i) printf("%d:%d:%2.1f%% ",-i+5,finger_stroke_distrib[i],100.0*finger_stroke_distrib[i]/total_finger);
	printf("\n R: "); for(i=6;i<10;++i) printf("%d:%d:%2.1f%% ",i-4,finger_stroke_distrib[i],100.0*finger_stroke_distrib[i]/total_finger);
	puts("\n");
	printf("Counted space strokes : %d\n", counted_space);
	printf("Untreated characters: %d\n", key_untreated_character_total[key_layout_num]);

	puts("");

	printf("Continuous one-hand strokes frequency distribution");
	for(i=0;i<2;++i) {
		printf("\n %c:",i==0?'L':'R');
		for(j=0;j<12;++j)
			if(continuous_same_hand_stroke[i][j]) printf(" %d:%d(%2.1f%%)",j+1,(j+1)*continuous_same_hand_stroke[i][j],100.0*(j+1)*continuous_same_hand_stroke[i][j]/(i==0 ? left_keystroke : right_keystroke));
	}
	puts("\n");

	printf("Continuous one-key strokes (L/R) : %d (%d/%d)\n",total_same_key_stroke, left_same_key_stroke, right_same_key_stroke);
	for(i=0;i<10;++i) {
		k=0;
		for(j=1;j<=6;++j)
			if(continuous_same_key_stroke[i][j]) k=1;
		if(!k) continue;
		k=0;
		printf(" %2d: ",i-5<0 ? i-5 : i-4);
		for(j=1;j<=6;++j) {
			if(continuous_same_key_stroke[i][j]) {
				printf("%d*%d",j,continuous_same_key_stroke[i][j]);
				k += continuous_same_key_stroke[i][j] * j;
			}
			if(j!=6 && continuous_same_key_stroke[i][j+1]) printf("+");
		}
		printf(" = %d\n",k);
	}
	puts("");
	printf("Total Continuous one-finger strokes : %d (L:%d R:%d)\n", left_same_finger_stroke+right_same_finger_stroke, left_same_finger_stroke, right_same_finger_stroke);
	int total_same_finger_stroke = 0;
	for(i=0;i<10;i++) total_same_finger_stroke += same_finger_stroke_distrib[i];
	for(i=0;i<10;++i) {
		k=0;
		for(j=1;j<=6;++j) 
			if(continuous_same_finger_stroke[i][j]) k=1;
		if(!k) continue;
		k=0;
		printf(" %2d: ",i-5<0 ? i-5 : i-4);
		for(j=1;j<=6;++j) {
			if(continuous_same_finger_stroke[i][j]) {
				printf("%d*%d",j,continuous_same_finger_stroke[i][j]);
				k+=continuous_same_finger_stroke[i][j] * j;
			}
			if(j!=6 && continuous_same_finger_stroke[i][j+1]) printf("+");
		}
		printf(" = %d (%2.1f%%)\n",k,100.0*same_finger_stroke_distrib[i]/total_same_finger_stroke);
	}
	printf("Continuous one-finger strokes involved in L/R shifts : %d/%d \n",left_shift_same_finger_stroke, right_shift_same_finger_stroke); // 윗글쇠(shift)가 일으키는 새끼 손가락 거듭치기 타수
	
	puts("");
	printf("Finger/Line frequency (Total key) :\n");	// 모든 글쇠 손가락-줄 타수 통계
	for(j=0;j<10;++j) {
		if(j==4 || j==5) continue;		
		if(j) printf("\t");
		if(j<4) printf("L%d", abs(j-5));		
		if(j>5) printf("R%d", j-4);
	}
	puts("");
	for(i=5;i>=2;--i) {
		for(j=0;j<10;++j) {
			if(j!=4 && j!=5)
				printf("%d\t",finger_line_freq[j][i]);
		}
		printf("\n");
	}
	puts("");
	printf("Finger/Line frequency (General key) :\n");	// 일반 글쇠 손가락-줄 타수 통계 (둘째 손가락 자리를 더 구분함)
	for(j=0;j<10;++j) {
		if(j) printf("\t");
		if(j<4) printf("L%d", abs(j-5));
		if(j==4) printf("L2-1", j-5);
		if(j==5) printf("R2-1", j-4);
		if(j>5) printf("R%d", j-4);
	}
	puts("");
	for(i=5;i>1;--i) {
		for(j=0;j<10;++j) {
			if(j) printf("\t");
			printf("%d",general_key_position_freq[j][i]);
		}
		printf("\n");
	}

	for(i=0;i<2;++i) free(continuous_same_hand_stroke[i]);
	for(i=0;i<10;++i) {
		free(continuous_same_finger_stroke[i]);
		free(continuous_same_key_stroke[i]);
	}
	free(continuous_same_hand_stroke);
	free(continuous_same_finger_stroke);
	free(continuous_same_key_stroke);
} // parse()
 
void parse_han(unsigned char *h,int *flag,int *cho,int *jung,int *jong)
{
	if(flag) *flag=h[0] & 128;
	if(cho)  *cho =(h[0] & (64+32+16+8+4))>>2;
	if(jung) *jung=((h[0] & (2+1))<<3)|((h[1] & (128+64+32))>>5);
	if(jong) *jong=(h[1] & (1+2+4+8+16));
}

/* 한글 자소 코드
 초성: NL(1) ㄱ(2) ㄲ(3) ㄴ(4) ㄷ(5) ㄸ(6) ㄹ(7) ㅁ(8) ㅂ(9) ㅃ(10) ㅅ(11) ㅆ(12)
	   ㅇ(13) ㅈ(14) ㅉ(15) ㅊ(16) ㅋ(17) ㅌ(18) ㅍ(19) ㅎ(20)
*/
char cho[21][3] = {
	"","",
	"ㄱ", "ㄲ", "ㄴ", "ㄷ", "ㄸ", "ㄹ", "ㅁ", "ㅂ", "ㅃ", "ㅅ", "ㅆ",
	"ㅇ", "ㅈ", "ㅉ", "ㅊ", "ㅋ", "ㅌ", "ㅍ", "ㅎ"
};
/*
 중성: NL(2) ㅏ(3) ㅐ(4) ㅑ(5) ㅒ(6) ㅓ(7) ㅔ(10) ㅕ(11) ㅖ(12) ㅗ(13) ㅘ(14)
	   ㅙ(15) ㅚ(18) ㅛ(19) ㅜ(20) ㅝ(21) ㅞ(22) ㅟ(23) ㅠ(26) ㅡ(27) ㅢ(28) ㅣ(29)
*/
char jung[30][3]={
	"", "", "",
	"ㅏ", "ㅐ", "ㅑ", "ㅒ", "ㅓ", "", "",
	"ㅔ", "ㅕ", "ㅖ", "ㅗ", "ㅘ",
	"ㅙ", "", "", "ㅚ",
	"ㅛ", "ㅜ", "ㅝ", "ㅞ", "ㅟ", "", "",
	"ㅠ", "ㅡ", "ㅢ", "ㅣ"
};
/*
종성: NL(1) ㄱ(2) ㄲ(3) ㄳ(4) ㄴ(5) ㄵ(6) ㄶ(7) ㄷ(8) ㄹ(9) ㄺ(10) ㄻ(11) ㄼ(12)
	   ㄽ(13) ㄾ(14) ㄿ(15) ㅀ(16) ㅁ(17) ㅂ(19) ㅄ(20) ㅅ(21) ㅆ(22) ㅇ(23) ㅈ(24)
	   ㅊ(25) ㅋ(26) ㅌ(27) ㅍ(28) ㅎ(29)
*/
char jong[30][3]={
	"","",
	"ㄱ","ㄲ","ㄳ","ㄴ","ㄵ","ㄶ","ㄷ","ㄹ","ㄺ","ㄻ","ㄼ",
	"ㄽ","ㄾ","ㄿ","ㅀ","ㅁ", "",
	"ㅂ","ㅄ","ㅅ","ㅆ","ㅇ","ㅈ","ㅊ","ㅋ","ㅌ","ㅍ","ㅎ"
};

// 1바이트 특수기호, 숫자 ! @ # $ % ^ & *
char one_byte_special[]={
	'0', '1', '2', '3', '4', '5', '6', '7', '8', '9',
	'`', '~',
	'!', '@', '#', '$', '%',
	'^', '&', '*', '(', ')',
	'-', '_', '=', '+',
	'[', ']', '{', '}', '\\', '|',
	';', ':', '\'', '\"',
	',', '.', '<',  '>', '/', '?'
};

char two_byte_special[][5] = {
	"·", "…", "〃", "―",
	"‘", "’", "“", "”", "※",
	"〈","〉", "《","》","「", "」", "『", "』", "【", "】",
	"±", "×", "÷", "≠",
	"°", "′", "″", "℃", "℉", "™",
	
	"○", "□",	"△", "▽", "◁", "▷",
	"→", "←", "↑", "↓", "↔",
	"☆", "♡", "☎", "☜", "☞",
	"￥", "€", "®", "©", "ⓒ"
};
unsigned char two_byte_special_code[][3] = {
	{0xD9,0x34}, {0xD9,0x36}, {0xD9,0x38}, {0xD9,0x3A},
	{0xD9,0x3E}, {0xD9,0x3F}, {0xD9,0x40}, {0xD9,0x41}, {0xD9,0x68},
	{0xD9,0x44}, {0xD9,0x45}, {0xD9,0x46}, {0xD9,0x47}, {0xD9,0x48}, {0xD9,0x49}, {0xD9,0x4A}, {0xD9,0x4B}, {0xD9,0x4C}, {0xD9,0x4D},
	{0xD9,0x4E}, {0xD9,0x4F}, {0xD9,0x50}, {0xD9,0x51},
	{0xD9,0x56}, {0xD9,0x57}, {0xD9,0x58}, {0xD9,0x59}, {0xD9,0xB5}, {0xD9,0xE2},
	
	{0xD9,0x6B}, {0xD9,0x70}, {0xD9,0x72}, {0xD9,0x74}, {0xD9,0xB7}, {0xD9,0xB9},
	{0xD9,0x76}, {0xD9,0x77}, {0xD9,0x78}, {0xD9,0x79}, {0xD9,0x7A},
	{0xD9,0x69}, {0xD9,0xBD}, {0xD9,0xCF}, {0xD9,0xD0}, {0xD9,0xD1},
	{0xD9,0x5D}, {0xD9,0xE6}, {0xD9,0xE7}, {0xDC,0xCF}, {0xDC,0xCF}
};

/* 닿소리 모음
NL(1) ㄱ(2) ㄲ(3) ㄳ(4) ㄴ(5) ㄵ(6) ㄶ(7) ㄷ(8) ㄸ(9) ㄹ(10) ㄺ(11) ㄻ(12) ㄼ(13) ㄽ(14) ㄾ(15) ㄿ(16) ㅀ(17) ㅁ(18)
ㅂ(19) ㅃ(20) ㅄ(21) ㅅ(22) ㅆ(23) ㅇ(24) ㅈ(25) ㅉ(26) ㅊ(27) ㅋ(28) ㅌ(29) ㅍ(30) ㅎ(31)
*/
char dah[][3] = {
	"", "ㄱ", "ㄲ", "ㄳ", "ㄴ", "ㄵ", "ㄶ", "ㄷ", "ㄸ", "ㄹ", "ㄺ", "ㄻ", "ㄼ", "ㄽ", "ㄾ", "ㄿ", "ㅀ", "ㅁ",
	"ㅂ", "ㅃ", "ㅄ", "ㅅ", "ㅆ", "ㅇ", "ㅈ", "ㅉ", "ㅊ", "ㅋ", "ㅌ", "ㅍ", "ㅎ"
};
char hot_dah[][3] = {
	"", "ㄱ", "ㄴ", "ㄷ", "ㄹ", "ㅁ",
	"ㅂ", "ㅅ", "ㅇ", "ㅈ", "ㅊ", "ㅋ", "ㅌ", "ㅍ", "ㅎ"
};
char dah_key[][3] = {
	"", "ㄱ", "ㄴ", "ㄷ", "ㄹ", "ㅁ", "ㅂ", "ㅅ", "ㅆ", "ㅇ", "ㅈ", "ㅊ", "ㅋ", "ㅌ", "ㅍ", "ㅎ"
};

/*
 중성: NL(2) ㅏ(3) ㅐ(4) ㅑ(5) ㅒ(6) ㅓ(7) ㅔ(10) ㅕ(11) ㅖ(12) ㅗ(13) ㅘ(14)
	   ㅙ(15) ㅚ(18) ㅛ(19) ㅜ(20) ㅝ(21) ㅞ(22) ㅟ(23) ㅠ(26) ㅡ(27) ㅢ(28) ㅣ(29)
*/
char hol[][3] = {
	"", "ㅏ", "ㅐ", "ㅑ", "ㅒ", "ㅓ", "", "", "ㅔ", "ㅕ", "ㅖ", "ㅗ", "ㅘ",
	"ㅙ", "", "", "ㅚ", "ㅛ", "ㅜ", "ㅝ", "ㅞ", "ㅟ", "", "", "ㅠ", "ㅡ", "ㅢ", "ㅣ"
};
char hot_hol[][3] = {
	"", "ㅏ", "ㅐ", "ㅑ", "ㅒ", "ㅓ", "ㅔ", "ㅕ", "ㅖ", "ㅗ",
	"ㅛ", "ㅜ", "ㅠ", "ㅡ", "ㅣ"
};
char hol_key[][3] = {
	"", "ㅏ", "ㅐ", "ㅑ", "ㅒ", "ㅓ", "ㅔ", "ㅕ", "ㅖ", "ㅗ",
	"ㅛ", "ㅜ", "ㅠ", "ㅡ", "ㅢ", "ㅣ"
};

int is_dahsori(char *ch) {
	int i;
	for(i=0; i<sizeof(dah)/sizeof(*dah); ++i)
		if(!strcmp(ch, dah[i])) return i;
	return NULL;
}

int is_holsori(char *hol_ch) {
	int i;
	for(i=0; i<sizeof(hol)/sizeof(*hol); ++i)
		if(!strcmp(hol_ch, hol[i])) return i;
	return NULL;
}

int dahsori_key(char *dah_ch) {
	int i;
	for(i=0; i<sizeof(dah_key)/sizeof(*dah_key); ++i)
		if(!strcmp(dah_ch, dah_key[i])) return i;
	return NULL;
}

int holsori_key(char *ch) {
	int i;
	for(i=0; i<sizeof(hol_key)/sizeof(*hol_key); ++i)
		if(!strcmp(ch, hol_key[i])) return i;
	return NULL;
}

int is_hot_dahsori(char *ch) {
	int i;
	for(i=0; i<sizeof(hot_dah)/sizeof(*hot_dah); ++i)
		if(!strcmp(ch, hot_dah[i])) return i;
	return NULL;
}

int is_hot_holsori(char *ch) {
	int i;
	for(i=0; i<sizeof(hot_hol)/sizeof(*hot_hol); ++i)
		if(!strcmp(ch, hot_hol[i])) return i;
	return NULL;
}

 
void main(int argc, char **argv)
{
	FILE *fp; unsigned char *buff; int f=0;
	int i,j,k,l;
	int choseong,jungseong,jongseong;
	int prev_choseong, prev_jungseong, prev_jongseong;
	int prev_key, next_key;

	int *one_byte_special_freq = (int *)calloc(sizeof(one_byte_special)/sizeof(char),sizeof(int));
	int *two_byte_special_freq = (int *)calloc(sizeof(two_byte_special)/sizeof(*two_byte_special),sizeof(int));

	int hangeul_natnae=0; // 한글 낱내 수
	int enter=0, space=0;	// 줄바꿈, 빈칸 수
	int total_digits=0;	// 숫자 잦기

	int *cho_freq=(int *)calloc(sizeof(cho)/sizeof(*cho), sizeof(int));
	int *jung_freq=(int *)calloc(sizeof(jung)/sizeof(*jung), sizeof(int));
	int *jong_freq=(int *)calloc(sizeof(jong)/sizeof(*jong), sizeof(int));

	int **jong_to_cho_consonant_comb_freq; // 닿소리 이음 잦기(결합 빈도) [끝소리][첫소리]
	jong_to_cho_consonant_comb_freq = (int **)malloc(sizeof(jong)/sizeof(*jong)*sizeof(int *));
	for(i=0;i<sizeof(jong)/sizeof(*jong);++i)
		jong_to_cho_consonant_comb_freq[i] = (int *)calloc(sizeof(cho)/sizeof(*cho),sizeof(int));
	
	int **jung_to_jong_consonant_comb_freq; // 가운뎃소리-끝소리 이음 잦기 [가운뎃소리][끝소리]
	jung_to_jong_consonant_comb_freq = (int **)malloc(sizeof(jung)/sizeof(*jung)*sizeof(int *));
	for(i=0;i<sizeof(jung)/sizeof(*jung);++i)
		jung_to_jong_consonant_comb_freq[i] = (int *)calloc(sizeof(jong)/sizeof(*jong),sizeof(int));

	int **dah_key_consonant_comb_freq; // 닿소리 글쇠 이음 잦기 [앞닿소리][뒷닿소리]
	dah_key_consonant_comb_freq = (int **)malloc(sizeof(int *)*sizeof(dah_key)/sizeof(*dah_key));
	for(i=0;i<sizeof(dah_key)/sizeof(*dah_key);++i)
		dah_key_consonant_comb_freq[i] = (int *)calloc((sizeof(dah_key))/sizeof(*dah_key),sizeof(int));

	int **hol_to_dah_key_consonant_comb_freq; // 가운뎃소리-끝소리 글쇠 이음 잦기 [가운뎃소리 글쇠][끝소리 글쇠]
	hol_to_dah_key_consonant_comb_freq = (int **)malloc(sizeof(int *)*sizeof(hol_key)/sizeof(*hol_key));
	for(i=0;i<sizeof(hol_key)/sizeof(*hol_key);++i)
		hol_to_dah_key_consonant_comb_freq[i] = (int *)calloc(sizeof(dah_key)/sizeof(*dah_key),sizeof(int));

	bool option=false;	// 선택 사항이 있는지 없는지
	bool option_beol2 = false;
	bool option_beol3 = false;
	bool option_alphabet_stat = false;
	bool option_combination_stat = false;

	for(i=0;i<argc;++i) {
		if(argv[i][0]=='/') {
			switch(argv[i][1]) {
				case '2' :	// 2벌 자판 통계
					option_beol2 = true;
					break;
				case '3' :	// 3벌 자판 통계
					option_beol3 = true;
					break;
				case 'j' :	// 잦기 통계
					option_alphabet_stat= true;
					break;
				case 'i' :	// 이음 잦기(결합 빈도) 통계
					option_combination_stat= true;
					break;
			}
			option = true;
		}
	}
	if(!option) {
		option_beol2 = true;
		option_beol3 = true;
		option_alphabet_stat = true;
		option_combination_stat= true;
	}
 
	fp=fopen((argc>=2 ? argv[1]:"sample.txt"),"rt");		// 분석할 글을 연다.
	if(fp==NULL) {puts("Cannot open sample file."); exit(1);}

	fseek(fp,0,SEEK_END); f=ftell(fp);
	printf("Reading %d bytes\n",f);
	buff=(unsigned char *)calloc(f,sizeof(unsigned char));
	fseek(fp,0,SEEK_SET);
	//fread(buff,f,sizeof(unsigned char),fp);

	f=0;
	while(!feof(fp)) {		// 파일의 끝을 다시 셈한다. (ftell 함수로 얻은 파일 크기는 든 내용보다 더 큼)
		*(buff+f) = getc(fp);
		++f;
	}
	fclose(fp);
	
	char **key_qwerty;
	int *key_pos;
	char str_qwerty[10], str_key[10], str[250], *ptr;
	char keyboard_layout_filename[300];

	char ***key_cho, ***key_jung, ***key_jong, ***key_one_byte_special, ***key_two_byte_special;
	int keyboard_layout_total = 0;

	_finddatai64_t c_file;
	intptr_t hFile;
	char path[] = "keyboard/*.txt";
	if ( (hFile = _findfirsti64(path, &c_file)) == -1L ) {
		switch(errno) {
			case ENOENT:
				printf("No keyboard layout file.\n"); break;
			case EINVAL:
				printf("Invalid path of keyboard layout file.\n"); break;
			case ENOMEM:
				printf("Not enough memory or keyboard layout file name too long.\n"); break;
			default:
				printf("Unknown error.\n"); break;
		}
	} else {
		key_layout_filenames = (char **)malloc(100*sizeof(char *));
		do {
			if(strcmp(c_file.name,"")) {
				++keyboard_layout_total;
				key_layout_filenames[keyboard_layout_total-1] = (char *)malloc((strlen(c_file.name)+1)*sizeof(char));
				strcpy(key_layout_filenames[keyboard_layout_total-1], c_file.name);
			}
		 } while(!_findnexti64(hFile, &c_file));
		 _findclose(hFile);
	} // end else

	key_cho = (char ***)calloc(keyboard_layout_total, sizeof(char **));
	key_jung = (char ***)calloc(keyboard_layout_total, sizeof(char **));
	key_jong = (char ***)calloc(keyboard_layout_total, sizeof(char **));
	key_one_byte_special = (char ***)calloc(keyboard_layout_total, sizeof(char **));
	key_two_byte_special = (char ***)calloc(keyboard_layout_total, sizeof(char **));

	key_qwerty = (char **)malloc(keyboard_layout_total*sizeof(char *));
	key_pos = (int *)calloc(keyboard_layout_total,sizeof(int));

	key_set_type = (int *)calloc(keyboard_layout_total,sizeof(int));
	key_fingering = (char **)calloc(keyboard_layout_total,sizeof(char *));
	key_untreated_character_total = (int *)calloc(keyboard_layout_total, sizeof(int));
	
	for(i=0;i<keyboard_layout_total;++i) {
		key_qwerty[i] = (char *)calloc(f*2,1);

		key_cho[i] = (char **)calloc(sizeof(cho)/sizeof(*cho), sizeof(char *));
		key_jong[i] = (char **)calloc(sizeof(jung)/sizeof(*jung), sizeof(char *));
		key_jung[i] = (char **)calloc(sizeof(jong)/sizeof(*jong), sizeof(char *));

		key_one_byte_special[i] = (char **)calloc(sizeof(one_byte_special)/sizeof(char), sizeof(char *));
		key_two_byte_special[i] = (char **)calloc(sizeof(two_byte_special)/sizeof(*two_byte_special), sizeof(char *));
		for(j=0;j<sizeof(cho)/sizeof(*cho);++j) key_cho[i][j] = (char *)calloc(5, sizeof(char));
		for(j=0;j<sizeof(jung)/sizeof(*jung);++j) key_jung[i][j] = (char *)calloc(5, sizeof(char));
		for(j=0;j<sizeof(jong)/sizeof(*jong);++j) key_jong[i][j] = (char *)calloc(5, sizeof(char));
		for(j=0;j<sizeof(one_byte_special)/sizeof(char);++j) key_one_byte_special[i][j] = (char *)calloc(5, sizeof(char));
		for(j=0;j<sizeof(two_byte_special)/sizeof(*two_byte_special);++j) key_two_byte_special[i][j] = (char *)calloc(5, sizeof(char));
	}

	for(i=0;i<keyboard_layout_total;++i) {
		strcpy(keyboard_layout_filename, "keyboard/");
		strcat(keyboard_layout_filename, key_layout_filenames[i]);
		fp=fopen(keyboard_layout_filename,"r");		// 자판 배열 파일을 연다.
		if(fp==NULL) {
			printf("Cannot open keyboard layout file: %s\n",keyboard_layout_filename);
			strcpy(key_layout_filenames[i],"");
			fclose(fp);
		} else {
			while(fgets(str,300,fp)) {		// 자판 파일을 한 줄씩 읽는다.	
				j = atoi(str);
				if(!j) {
					ptr = strtok(str,":");		// :을 기준으로 문자열을 끊는다.
					if(str[0]=='/' && str[1]=='/') continue;	// 주석문
					else if(!stricmp(ptr,"set type")) {
						ptr = strtok(NULL,"\n");
						key_set_type[i] = atoi(ptr);
						continue;
					} else if (!stricmp(ptr,"fingering")) {
						ptr = strtok(NULL,"\n");
						key_fingering[i] = (char *)malloc(strlen(ptr)*sizeof(char));
						l=0;
						for(k=0; k<strlen(ptr); ++k) {
							if(ptr[k] != ' ') key_fingering[i][l++] = ptr[k];
						}
						key_fingering[i][l] = '\0';
						
						// 운지법을 지정하지 않았거나 분석기에서 지원하지 않는 운지법을 지정했다면 기본 운지법(쿼티 자판식)으로 분석한다.
						if(!stricmp(key_fingering[i],"")) strcpy(key_fingering[i],"Qwerty");
						else if(stricmp(key_fingering[i],"Pyojun") && stricmp(key_fingering[i],"Gong") && stricmp(key_fingering[i],"Gong3"))
							strcpy(key_fingering[i],"Unknown");

						if(key_fingering[i][0]>='a' && key_fingering[i][0]<='z') key_fingering[i][0]-=32;
						continue;
					}
				}
				ptr = strtok(str,"\t");		// 탭으로 나뉜 문자열을 끊는다.
				if(!ptr || !j) continue;
				
				ptr = strtok(NULL,"\t");
				if(!ptr) continue;
				strcpy(str_key, ptr);		// str_key : 낱소리, 기호에 대응하는 글쇠 

				ptr = strtok(NULL,"\n");
				if(!ptr) continue;
				strcpy(str_qwerty, ptr);		// str_qwerty : 쿼티 자판에 대응하는 글쇠

				switch(j) {
					case 1:		// 한글 첫소리일 때
						for(k=0;k<sizeof(cho)/sizeof(*cho);++k)
							if(!strcmp(str_key, cho[k])) break;
						if(k!=sizeof(cho)/sizeof(*cho)) {strcpy(key_cho[i][k], str_qwerty);}
						break;
					case 2:		// 한글 가운뎃소리일 때
						for(k=0;k<sizeof(jung)/sizeof(*jung);++k)
							if(!strcmp(str_key, jung[k])) break;
						if(k!=sizeof(jung)/sizeof(*jung)) strcpy(key_jung[i][k], str_qwerty);
						break;
					case 3:		// 한글 끝소리일 때
						for(k=0;k<sizeof(jong)/sizeof(*jong);++k)
							if(!strcmp(str_key, jong[k])) break;
						if(k!=sizeof(jong)/sizeof(*jong)) strcpy(key_jong[i][k], str_qwerty);
						break;
					case 4:		// 숫자, 기호일 때 (한글이 아닐 때)
						if(strlen(str_key)==1) {		// 1byte 문자일 때
							for(k=0;k<sizeof(one_byte_special)/sizeof(char);++k)
								if(str_key[0] == one_byte_special[k]) break;
							if(k!=sizeof(one_byte_special)/sizeof(char)) strcpy(key_one_byte_special[i][k], str_qwerty);
						} else {	// 2byte 문자일 때
							for(k=0;k<sizeof(two_byte_special)/sizeof(*two_byte_special);++k)
								if(!strcmp(str_key, two_byte_special[k])) break;
							if(k!=sizeof(two_byte_special)/sizeof(*two_byte_special)) strcpy(key_two_byte_special[i][k], str_qwerty);
						}
						break;
				}
			}
			fclose(fp);
		}
	}
	
	for(i=0,j=strlen((char *)buff);i<j;i++) {
		if(buff[i]>=128) {

			// 2바이트 특수기호인지 알아보고 특수기호를 셈한다.
			for(k=0; k<sizeof(two_byte_special)/sizeof(*two_byte_special); ++k) {
				if(buff[i]==two_byte_special_code[k][0] && buff[i+1]==two_byte_special_code[k][1]) { // 문자 목록에 있는 2바이트 특수기호일 때
					++two_byte_special_freq[k];
					for(l=0; l<keyboard_layout_total; ++l) {
						if(strcmp(key_two_byte_special[l][k],"")) {		// 자판 배열에 있는 2바이트 특수기호일 때
							if(!strcmp(two_byte_special[k], "‘")) {		// 열고 닫는 따옴표를 일반 자판에 있는 따옴표로 치는 것으로 한다.
								strcpy(key_qwerty[l]+key_pos[l], "'");
								++key_pos[l];
							} else if(!strcmp(two_byte_special[k], "’")) {
								strcpy(key_qwerty[l]+key_pos[l], "'");
								++key_pos[l];
							} else if(!strcmp(two_byte_special[k], "“")) {
								strcpy(key_qwerty[l]+key_pos[l], "\"");
								++key_pos[l];
							} else if(!strcmp(two_byte_special[k], "”")) {
								strcpy(key_qwerty[l]+key_pos[l], "\"");
								++key_pos[l];
							} else {
								strcpy(key_qwerty[l]+key_pos[l], key_two_byte_special[l][k]);
								key_pos[l] += strlen(key_two_byte_special[l][k]);
							}
						} else {		// 자판 배열에 없는 2바이트 특수기호일 때
							*(key_qwerty[l]+key_pos[l]) = 0x18;		// 자판 배열에 없는 문자를 쓰지 않는 문자로 채둔다.
							++key_pos[l];
							++key_untreated_character_total[l];
						}
					}
					break;
				}
			}
			if(k!=sizeof(two_byte_special)/sizeof(*two_byte_special) || is_hanja(buff[i],buff[i+1])) {
				prev_choseong = prev_jungseong = prev_jongseong = NULL;
				i++;
				continue;
			}

			parse_han(buff+i, NULL, &choseong, &jungseong, &jongseong);

			if(choseong>=sizeof(cho)/sizeof(*cho) || jungseong>=sizeof(jung)/sizeof(*jung) || jongseong>=sizeof(jong)/sizeof(*jong)) {
				prev_choseong = prev_jungseong = prev_jongseong = NULL;
				i++;
				continue;
			}

			hangeul_natnae++;
			++cho_freq[choseong]; ++jung_freq[jungseong], ++jong_freq[jongseong];

			
			if(jongseong) {	// 끝소리가 있을 때에 가운뎃소리와 끝소리의 이음 잦기를 셈한다.
				if(holsori_key(jung[jungseong])) prev_key = holsori_key(jung[jungseong]);
				else if(!strcmp(jung[jungseong],"ㅘ")) prev_key = holsori_key("ㅏ");
				else if(!strcmp(jung[jungseong],"ㅙ")) prev_key = holsori_key("ㅐ");
				else if(!strcmp(jung[jungseong],"ㅚ")) prev_key = holsori_key("ㅣ");
				else if(!strcmp(jung[jungseong],"ㅝ")) prev_key = holsori_key("ㅓ");
				else if(!strcmp(jung[jungseong],"ㅞ")) prev_key = holsori_key("ㅔ");
				else if(!strcmp(jung[jungseong],"ㅟ")) prev_key = holsori_key("ㅟ");
				else if(!strcmp(jung[jungseong],"ㅢ")) prev_key = holsori_key("ㅣ");
				else prev_key = 0;

				if(dahsori_key(jong[jongseong])) next_key = dahsori_key(jong[jongseong]);
				else if(!strcmp(jong[jongseong],"ㄲ")) next_key = is_holsori("ㄱ");
				//else if(!strcmp(jong[jongseong],"ㄸ")) next_key = dahsori_key("ㄷ");
				//else if(!strcmp(jong[jongseong],"ㅃ")) next_key = dahsori_key("ㅂ");
				else if(!strcmp(jong[jongseong],"ㅆ")) next_key = dahsori_key("ㅅ");
				else if(!strcmp(jong[jongseong],"ㅉ")) next_key = dahsori_key("ㅈ");
				else if(!strcmp(jong[jongseong],"ㄳ")) next_key = dahsori_key("ㄱ");
				else if(!strcmp(jong[jongseong],"ㄵ")) next_key = dahsori_key("ㄴ");
				else if(!strcmp(jong[jongseong],"ㄶ")) next_key = dahsori_key("ㄴ");
				else if(!strcmp(jong[jongseong],"ㄺ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㄻ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㄼ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㄽ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㄾ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㄿ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㅀ")) next_key = dahsori_key("ㄹ");
				else if(!strcmp(jong[jongseong],"ㅄ")) next_key = dahsori_key("ㅂ");
				else next_key = 0;
				
				++hol_to_dah_key_consonant_comb_freq[prev_key][next_key];
				++jung_to_jong_consonant_comb_freq[jungseong][jongseong];
			}
			
			if(jongseong && !is_hot_dahsori(jong[jongseong])) {	// 겹받침의 앞뒤 닿소리 글쇠 이음 잦기를 셈한다. (2벌식)
				if(dahsori_key(jong[jongseong])) { prev_key = NULL; next_key = NULL;}
				//else if(!strcmp(jong[jongseong],"ㄲ")) { prev_key = dahsori_key("ㄱ"); next_key = dahsori_key("ㄱ");}
				//else if(!strcmp(jong[jongseong],"ㄸ")) { prev_key = NULL; next_key = NULL;}
				//else if(!strcmp(jong[jongseong],"ㅃ")) { prev_key = NULL; next_key = NULL;}
				//else if(!strcmp(jong[jongseong],"ㅆ")) { prev_key = NULL; next_key = NULL;}
				//else if(!strcmp(jong[jongseong],"ㅉ")) { prev_key = NULL; next_key = NULL;}
				else if(!strcmp(jong[jongseong],"ㄳ")) { prev_key = dahsori_key("ㄱ"); next_key = dahsori_key("ㅅ");}
				else if(!strcmp(jong[jongseong],"ㄵ")) { prev_key = dahsori_key("ㄴ"); next_key = dahsori_key("ㅈ");}
				else if(!strcmp(jong[jongseong],"ㄶ")) { prev_key = dahsori_key("ㄴ"); next_key = dahsori_key("ㅎ");}
				else if(!strcmp(jong[jongseong],"ㄺ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㄱ");}
				else if(!strcmp(jong[jongseong],"ㄻ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅁ");}
				else if(!strcmp(jong[jongseong],"ㄼ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅂ");}
				else if(!strcmp(jong[jongseong],"ㄽ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅅ");}
				else if(!strcmp(jong[jongseong],"ㄾ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅌ");}
				else if(!strcmp(jong[jongseong],"ㄿ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅍ");}
				else if(!strcmp(jong[jongseong],"ㅀ")) { prev_key = dahsori_key("ㄹ"); next_key = dahsori_key("ㅎ");}
				else if(!strcmp(jong[jongseong],"ㅄ")) { prev_key = dahsori_key("ㅂ"); next_key = dahsori_key("ㅅ");}
				else { prev_key = NULL; next_key = NULL;}

				if(prev_key && next_key) ++dah_key_consonant_comb_freq[prev_key][next_key];
			}
			
			if(prev_jongseong && strcmp(jong[prev_jongseong],"")) {	// 앞 낱내의 끝소리와 현재 낱내의 첫소리와의 글쇠 이음 잦기를 셈한다. (2벌식)
				if(dahsori_key(jong[prev_jongseong])) prev_key = dahsori_key(jong[prev_jongseong]);
				else if(!strcmp(jong[prev_jongseong],"ㄲ")) prev_key = dahsori_key("ㄱ");
				else if(!strcmp(jong[prev_jongseong],"ㅆ")) prev_key = dahsori_key("ㅅ");
				else if(!strcmp(jong[prev_jongseong],"ㄳ")) prev_key = dahsori_key("ㅅ");
				else if(!strcmp(jong[prev_jongseong],"ㄵ")) prev_key = dahsori_key("ㅈ");
				else if(!strcmp(jong[prev_jongseong],"ㄶ")) prev_key = dahsori_key("ㅎ");
				else if(!strcmp(jong[prev_jongseong],"ㄺ")) prev_key = dahsori_key("ㄱ");
				else if(!strcmp(jong[prev_jongseong],"ㄻ")) prev_key = dahsori_key("ㅁ");
				else if(!strcmp(jong[prev_jongseong],"ㄼ")) prev_key = dahsori_key("ㅂ");
				else if(!strcmp(jong[prev_jongseong],"ㄽ")) prev_key = dahsori_key("ㅅ");
				else if(!strcmp(jong[prev_jongseong],"ㄾ")) prev_key = dahsori_key("ㅌ");
				else if(!strcmp(jong[prev_jongseong],"ㄿ")) prev_key = dahsori_key("ㅍ");
				else if(!strcmp(jong[prev_jongseong],"ㅀ")) prev_key = dahsori_key("ㅎ");
				else if(!strcmp(jong[prev_jongseong],"ㅄ")) prev_key = dahsori_key("ㅅ");
				else prev_key = NULL;
					
				if(dahsori_key(cho[choseong])) next_key = dahsori_key(cho[choseong]);
				else if(!strcmp(cho[choseong],"ㄲ")) next_key = dahsori_key("ㄱ");
				else if(!strcmp(cho[choseong],"ㄸ")) next_key = dahsori_key("ㄷ");
				else if(!strcmp(cho[choseong],"ㅃ")) next_key = dahsori_key("ㅂ");
				else if(!strcmp(cho[choseong],"ㅆ")) next_key = dahsori_key("ㅅ");
				else if(!strcmp(cho[choseong],"ㅉ")) next_key = dahsori_key("ㅈ");
				else next_key = NULL;

				if(prev_key && next_key) ++dah_key_consonant_comb_freq[prev_key][next_key];
				++jong_to_cho_consonant_comb_freq[prev_jongseong][choseong];
			}
			
			for(k=0; k<keyboard_layout_total; ++k) {			// 사용자 정의 자판 배열의 한글 낱소리들을 쿼티 배열로 옮긴다.
				if(!strcmp(key_layout_filenames[k],"")) continue;
				strcpy(key_qwerty[k]+key_pos[k], key_cho[k][choseong]); key_pos[k]+=strlen(key_cho[k][choseong]);
				strcpy(key_qwerty[k]+key_pos[k], key_jung[k][jungseong]); key_pos[k]+=strlen(key_jung[k][jungseong]);
				strcpy(key_qwerty[k]+key_pos[k], key_jong[k][jongseong]); key_pos[k]+=strlen(key_jong[k][jongseong]);
			}
			
			prev_choseong = choseong;
			prev_jungseong = jungseong;
			prev_jongseong = jongseong;
			i++;

		} else {		// 1바이트 문자일 때
			if(buff[i]==' ' || buff[i]=='\n') {		// 빈칸이거나 줄이 바뀔 때
				if(buff[i] == ' ') ++space;
				if(buff[i] == '\n') ++enter;
				for(l=0; l<keyboard_layout_total; ++l) {
					*(key_qwerty[l]+key_pos[l]) = buff[i]; ++key_pos[l];
				}

				prev_choseong = prev_jungseong = prev_jongseong = NULL;
				continue;
			}

			for(k=0;k<sizeof(one_byte_special)/sizeof(char);++k) 
				if(one_byte_special[k] == buff[i]) break;		// 1바이트 기호 및 숫자 목록에 있는지 판별한다.

			if(k == sizeof(one_byte_special)/sizeof(char)) {
				prev_choseong = prev_jungseong = prev_jongseong = NULL;
				continue;
			}

			if(k!=sizeof(one_byte_special)/sizeof(char)) {
				for(l=0; l<keyboard_layout_total; ++l) {
					if(strcmp(key_one_byte_special[l][k],"")) {
						strcpy(key_qwerty[l]+key_pos[l], key_one_byte_special[l][k]);
						key_pos[l] += strlen(key_one_byte_special[l][k]);
					} else {
						*(key_qwerty[l]+key_pos[l]) = 0x18;
						++key_pos[l];
						++key_untreated_character_total[l];
					}
				}
			}
			
			++one_byte_special_freq[k];
			if(one_byte_special[k] >= '0' && one_byte_special[k] <= '9') ++total_digits;

			prev_choseong = NULL;
			prev_jungseong = NULL;
			prev_jongseong = NULL;
		}
	}

	if(option_alphabet_stat) {
		int total_cho=0, total_jung=0, total_jong=0, total_signs=0;
		for(i=0;i<sizeof(one_byte_special)/sizeof(char);++i) total_signs += one_byte_special_freq[i];
		total_signs -= total_digits;
		for(i=0;i<sizeof(two_byte_special)/sizeof(*two_byte_special);++i) total_signs += two_byte_special_freq[i];
		for(i=2;i<sizeof(cho)/sizeof(*cho);++i) if(cho[i][0]) total_cho+=cho_freq[i];
		for(i=3;i<sizeof(jung)/sizeof(*jung);++i) if(jung[i][0]) total_jung+=jung_freq[i];
		for(i=2;i<sizeof(jong)/sizeof(*jong);++i) if(jong[i][0]) total_jong+=jong_freq[i];
		puts("");
		printf("Hangeul Natnae: %d\n",hangeul_natnae);
		printf(" (Cheot-sori: %d, Gaunget-sori: %d, Ggeut-sori: %d)\n",total_cho, total_jung, total_jong);
		puts("");
		printf("Digits: %d\n",total_digits);
		printf("Signs: %d\n",total_signs);
		printf("Enter : %d\n",enter);
		printf("Space : %d\n",space);
		puts("");

		printf("Cheot-sori Frequency: %d\n",total_cho);
		for(i=2; i<sizeof(cho)/sizeof(*cho); ++i) {
			if(strcmp(cho[i],"") || cho_freq[i])
			printf("%s:%9d (%2.2f%%)\n",cho[i], cho_freq[i], 100.0*cho_freq[i]/total_cho);
		}
		puts("");
		printf("Gaundet-sori Frequency: %d\n",total_jung);
		for(i=3; i<sizeof(jung)/sizeof(*jung); ++i) {
			if(strcmp(jung[i],"") || jung_freq[i])
				printf("%s:%9d (%2.2f%%)\n",jung[i], jung_freq[i], 100.0*jung_freq[i]/total_jung);
		}
		puts("");
		printf("Ggeut-sori Frequency: %d\n",total_jong);
		j=0;
		for(i=2; i<sizeof(jong)/sizeof(*jong); ++i) {
			if(strcmp(jong[i],"") || jong_freq[i])
			printf("%s:%9d (%2.2f%%)\n",jong[i], jong_freq[i], 100.0*jong_freq[i]/total_jong);
		}
		puts("");

		printf("Digits: %d\n",total_digits);
		for(i=0; i<10; ++i) printf(" %d:%d ",i,one_byte_special_freq[i]);
		puts("\n");
		for(i=10; i<sizeof(one_byte_special)/sizeof(char); ++i)
			printf("%c : %d\n", one_byte_special[i], one_byte_special_freq[i]);		// 숫자, 특수기호의 잦기를 출력한다.
		for(i=0; i<sizeof(two_byte_special)/sizeof(*two_byte_special); ++i)
			printf("%s : %d\n" ,two_byte_special[i], two_byte_special_freq[i]);		// 2바이트 특수기호(·, ※ 따위)의 잦기를 출력한다.

	}

	if(option_combination_stat) {
		printf("\nGgeutsori-Cheotsori ieum\n");		// 끝소리-첫소리 이음 잦기 
		printf("**");
		for(j=2;j<sizeof(cho)/sizeof(*cho);++j) printf(" %s",cho[j]);		
		printf("\n");
		for(i=2;i<sizeof(jong)/sizeof(*jong);++i) {
			printf("%s",jong[i]);
			for(j=2;j<sizeof(cho)/sizeof(*cho);++j) {
				printf(" %d",jong_to_cho_consonant_comb_freq[i][j]);
			}
			printf("\n");
		}
		puts("");

		printf("Dahsori Key Combination Frequncy\n");	// 닿소리 글쇠 결합 빈도 (2벌식 + ㅆ 받침)
		printf("*");
		for(j=1;j<sizeof(dah_key)/sizeof(*dah_key);++j) //if(dahsori_key(dah[j]))
			printf(" %s",dah_key[j]);
		printf("\n");
		for(i=1;i<sizeof(dah_key)/sizeof(*dah_key);++i) {
			printf("%s",dah_key[i]);
			for(j=1;j<sizeof(dah_key)/sizeof(*dah_key);++j)
				printf(" %d",dah_key_consonant_comb_freq[i][j]);
			printf("\n");
		}
		puts("");

		printf("Dahsori key combination frequency (no order)\n");
		printf("x");	// 닿소리 글쇠 결합 빈도 (2벌식 + ㅆ 받침, 앞뒤 차례 없이)
		for(j=1;j<sizeof(dah_key)/sizeof(*dah_key);++j)
			printf(" %s",dah_key[j]);
		printf("\n");
		for(i=1;i<sizeof(dah_key)/sizeof(*dah_key);++i) {
			printf("%s",dah_key[i]);
			for(j=1;j<sizeof(dah_key)/sizeof(*dah_key);++j)
				if(i==j) printf(" %d",dah_key_consonant_comb_freq[i][j]);
				else printf(" %d",dah_key_consonant_comb_freq[i][j]+dah_key_consonant_comb_freq[j][i]);
				printf("\n");
		}
		puts("");
		
		printf("Gaunde-Ggeut sori combination frequency\n");		// 가운뎃소리-끝소리 이음 잦기
		printf("++");
		for(j=3;j<sizeof(jung)/sizeof(*jung);++j)
			if(strcmp(jung[j],"")) 	printf(" %s",jung[j]);
		printf("\n");
		for(i=2;i<sizeof(jong)/sizeof(*jong);++i) {
			printf("%s",jong[i]);
			for(j=3;j<sizeof(jung)/sizeof(*jung);++j) {
				if(!strcmp(jung[j],"")) continue;
				printf(" %d",jung_to_jong_consonant_comb_freq[j][i]);
			}
			printf("\n");
		}
		puts("");
		
		printf("Hol-Dah sori key combination frequency\n");			// 가운뎃소리-끝소리 글쇠 이음 잦기
		printf("+");
		for(j=1;j<sizeof(hol_key)/sizeof(*hol_key);++j) printf(" %s",hol_key[j]);
		printf("\n");
		for(i=1;i<sizeof(dah_key)/sizeof(*dah_key);++i) {
			printf("%s",dah_key[i]);
			for(j=1;j<sizeof(hol_key)/sizeof(*hol_key);++j) {
				printf(" %d",hol_to_dah_key_consonant_comb_freq[j][i]);
			}
			printf("\n");
		}
		puts("");
	} // if(option_combination_stat)

	if(option_beol2) {
		for(i=0; i<keyboard_layout_total; ++i) {
			if(!strcmp(key_layout_filenames[i],"") || key_set_type[i]!=2) continue;

			printf("\n*** Result of %s (%d Sets, %s Fingering) ***\n\n",strtok(key_layout_filenames[i],"."),key_set_type[i],key_fingering[i]);
			init_type_info(i); parse(key_qwerty[i], i);
		}
	}
	if(option_beol3) {
		for(i=0; i<keyboard_layout_total; ++i) {
			if(!strcmp(key_layout_filenames[i],"") || key_set_type[i]!=3) continue;
			printf("\n*** Result of %s (%d Sets, %s Fingering) ***\n\n",strtok(key_layout_filenames[i],"."),key_set_type[i],key_fingering[i]);
			init_type_info(i); parse(key_qwerty[i], i);
		}
	}

	free(one_byte_special_freq); free(two_byte_special_freq);
	free(cho_freq);	free(jung_freq); free(jong_freq);

	for(i=0;i<sizeof(jong)/sizeof(*jong);++i)
		free(jong_to_cho_consonant_comb_freq[i]);
	free(jong_to_cho_consonant_comb_freq);

	for(i=0;i<sizeof(jung)/sizeof(*jung);++i)
		free(jung_to_jong_consonant_comb_freq[i]);
	free(jung_to_jong_consonant_comb_freq);

	for(i=0;i<sizeof(dah_key)/sizeof(*dah_key);++i)
		free(dah_key_consonant_comb_freq[i]);
	free(dah_key_consonant_comb_freq);

	for(i=0;i<sizeof(hol_key)/sizeof(*hol_key);++i)
		free(hol_to_dah_key_consonant_comb_freq[i]);
	free(hol_to_dah_key_consonant_comb_freq);

	for(i=0; i<keyboard_layout_total; ++i)
		free(key_layout_filenames[i]);
	free(key_layout_filenames);

	free(key_fingering);
	free(buff);

} // main()